/*
 *  Copyright (c) 2012 Silicon Graphics, Inc.  All Rights Reserved.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation; either version 2.1 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 */
#ifndef __GRU_MESQ_INTERNALS_H_
#define __GRU_MESQ_INTERNALS_H_

/* Status values for gru_send_message() */
#define MQE_OK			0	/* message sent successfully */
#define MQE_CONGESTION		1	/* temporary failure caused by network
					   congestion, try again */
#define MQE_QUEUE_FULL		2	/* queue is full and receiver not
					   keeping up */
#define MQE_PAGE_OVERFLOW	3	/* Queue overflowed a page. User bug */
#define MQE_BUG			4	/* unexpected library bug encountered */
#define MQE_BUG_TOO_BIG		10	/* Message too big */
#define MQE_AGAIN		-2	/* (internal) queue full sucessfully
					   handled */
#define MQE_UNEXPECTED_CB_ERR	-1	/* unexpected CB error. See CB for
					   details */
#define MQE_BUG_NO_RESOURCES	11	/* BUG - could not alloc GRU cb/dsr */

/* Value of the "present" flags in cachelines 1 & 2 of a message */
#define MQS_EMPTY               0
#define MQS_MES1                1
#define MQS_MES2                2
#define MQS_LINE2               3
#define MQS_NOOP                4

#define __gru_cacheline_aligned__				\
	__attribute__((__aligned__(GRU_CACHE_LINE_BYTES)))

#define HSTATUS(mq, h)  ((mq) + offsetof(struct message_queue, hstatus[h]))

/* Bit numbers for accessing sw_reserved0 bit 0 in DW1 of the CBR */
#define SW_RESERVED0_LOCK_BIT_DW1    		32

/* sw_reserved0 bit values used by library & kernel */
#define SW_RESERVED0_LOCK	    		0x01UL
#define SW_RESERVED0_ASYNC_MESQ_INSTRUCTION	0x02UL
#define SW_RESERVED0_FATAL_EXCEPTION		0x80UL

/*
 * Used for library MESQ synchronous instructions, ie. instruction
 * is immediately followed by a wait. There is no advantage in have the kernel
 * handle exceptions.
 */
#define SW_RESERVED0_USER_MESQ_SYNC	(SW_RESERVED0_LOCK)

/* Used for library MESQ asynchronous instructions, ie. lbrary does NOT
 * immediately wait for a completion. Kernel should attempt to handle
 * any exceptions. No user mode action is needed to guarantee the instruction
 * will complete.
 */
#define SW_RESERVED0_USER_MESQ_ASYNC	(SW_RESERVED0_LOCK | SW_RESERVED0_ASYNC_MESQ_INSTRUCTION)

/*
 * Used by the library MESQ exception handlers for handling ALL
 * MESQ exceptions. This include GAM instruction used for exception
 * recovery.
 */
#define SW_RESERVED0_USER_MESQ_EXCEPTION	(SW_RESERVED0_LOCK)

/*
 * Used for kernel MESQ-related instructions. Basically preserve the LOCK
 * bit to inhibit the library from seeing the exception while it's being handled
 * by the kernel.
 */
#define SW_RESERVED0_KERN_MESQ		(SW_RESERVED0_LOCK | SW_RESERVED0_ASYNC_MESQ_INSTRUCTION)

/*
 * IMA values for kernel & library.
 */
#define IMA_USER_MESQ_SYNC	(SW_RESERVED0_USER_MESQ_SYNC << IMA_SW_SHIFT)
#define IMA_USER_MESQ_ASYNC	(SW_RESERVED0_USER_MESQ_ASYNC << IMA_SW_SHIFT)
#define IMA_USER_MESQ_EXCEPTION	(SW_RESERVED0_USER_MESQ_EXCEPTION << IMA_SW_SHIFT)
#define IMA_KERN_MESQ		(SW_RESERVED0_KERN_MESQ << IMA_SW_SHIFT)

/* Macros for build q-head values for the first & second halves of a queue */
#define MESQ_HEAD0(ql)		gru_mesq_head(MESQ_HEAD_LINES, MESQ_HEAD_LINES + (ql) / 2)
#define MESQ_HEAD1(ql)		gru_mesq_head(MESQ_HEAD_LINES + (ql) / 2, MESQ_HEAD_LINES + (ql))

/* message queue header */
struct message_queue {
	union gru_mesqhead	head __gru_cacheline_aligned__;		/* CL 0 */
	long 			hstatus[2];
	unsigned long		qlines;
	char			mq_allocated;
	void 			*fill __gru_cacheline_aligned__;	/* CL 1 */
	void 			*next __gru_cacheline_aligned__;	/* CL 2 */
	void 			*limit;
	void 			*start;
	void 			*start2;
	char			data[0] __gru_cacheline_aligned__;	/* CL 3 */
};
#define MESQ_HEAD_LINES		(sizeof(struct message_queue) / GRU_CACHE_LINE_BYTES)

/* First word in every message - used by mesq interface */
struct message_header {
	char	present;
	char	present2;
	char 	fill1;
	char	fill2;
};

/*
 * Function for decoding MESQ instructions.
 */
static inline unsigned long gru_get_cb_tri0(void *cb)
{
	struct gru_instruction *ins = (void *)cb;
	return (ins->tri0 & 0x3fffff);
}

static inline unsigned long gru_get_cb_paddr(void *cb)
{
	struct gru_instruction *ins = (void *)cb;
	return ins->baddr0;
}

static inline void gru_restore_cb_paddr(void *cb, unsigned long val)
{
	struct gru_instruction *ins = (void *)cb;
	ins->baddr0 = val;
}

static inline void *gru_get_cb_data_pointer(void *cb)
{
	void *gseg = (void *)((unsigned long)cb & ~0xffff);
	return gru_get_data_pointer(gseg, gru_get_cb_tri0(cb) / GRU_CACHE_LINE_BYTES);
}

static inline int gru_get_cb_nelem(void *cb)
{
	struct gru_instruction *ins = (void *)cb;
	return ins->nelem;
}

/*
 * Save/restore the "present" flag that is in the second line of 2-line
 * messages
 */
static inline int get_present2(void *p)
{
	struct message_header *mhdr = p + GRU_CACHE_LINE_BYTES;
	return mhdr->present;
}

static inline void set_present2(void *p, int val)
{
	struct message_header *mhdr = p + GRU_CACHE_LINE_BYTES;
	mhdr->present = val;
}

static inline int gru_is_async_mesq_library_instruction(void *cb)
{
	struct gru_instruction_bits *ins = (void *)cb;

	return !!(ins->sw_reserved0 & SW_RESERVED0_ASYNC_MESQ_INSTRUCTION);

}

/*
 * Clear the MESQ instruction lock in the CBR of a user instruction.
 */
static inline void gru_unlock_mesq_instruction(void *cb)
{
	struct gru_instruction *ins = (void *)cb;
	if (!(ins->sw_reserved0 & SW_RESERVED0_LOCK)) {
		unsigned long *p = cb;
		printk("ZZZZZZZZZZZ kernel not locked: cb %p (0x%lx 0x%lx)\n", cb, p[0], p[1]);
		BUG();
	}
	clear_bit(SW_RESERVED0_LOCK_BIT_DW1, ((unsigned long*)(cb)));
}

/*
 * Trylock the MESQ instruction lock in the CBR of a user instruction.
 */
static inline int gru_trylock_mesq_instruction(void *cb)
{
	 cb += 8;
	 return !test_and_set_bit(SW_RESERVED0_LOCK_BIT_DW1, ((unsigned long*)(cb)));
}

extern int gru_send_message_failure(void *cb, struct gru_message_queue_desc *mq);
extern char tlb_mesq_mode_kernel;

#endif 		/* __GRU_MESQ_INTERNALS_H_ */
