/*
 *  Copyright (c) 2008-2013 Silicon Graphics, Inc.  All Rights Reserved.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation; either version 2.1 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 */

#ifndef __GRU_H__
#define __GRU_H__

/*
 * GRU architectural definitions
 */
#define GRU_CACHE_LINE_BYTES		64
#define GRU_HANDLE_STRIDE		256
#define GRU_CB_BASE			0
#define GRU_DS_BASE			0x20000

/*
 * Minimum page size used to map a GRU context
 */
#define GRU_MIN_GSEG_PAGESIZE	(256 * 1024UL)

/*
 * Statictics kept for each context.
 */
struct gru_gseg_statistics {
	unsigned long	fmm_tlbmiss;
	unsigned long	upm_tlbmiss;
	unsigned long	tlbdropin;
	unsigned long	context_stolen;
	unsigned long	interrupts;
	unsigned long	reserved[9];
};

/* Flags for GRU options on the gru_create_context() call */
/* Select one of the follow 5 options to specify how TLB misses are handled */
#define GRU_OPT_MISS_DEFAULT			0x0000	/* Use default mode */
#define GRU_OPT_MISS_USER_POLL			0x0001	/* User will poll CB for faults */
#define GRU_OPT_MISS_FMM_INTR			0x0002	/* Send interrupt to handle fault */
#define GRU_OPT_MISS_FMM_INTR_NO_POLL		0x0003	/* Send interrupt to handle fault - no user UPM polling  */
#define GRU_OPT_MISS_FMM_INTR_NO_POLL_MESQ	0x0004	/* Above + mesq exceptions handled by the kernel */
#define GRU_OPT_MISS_MASK			0x0007	/* Mask for TLB MISS option */

/*
 * Basic types to improve type checking. User apps should never directly
 * reference fields within these structures.
 */
typedef struct {unsigned long cookie; } gru_cookie_t;
typedef struct gru_segment_s gru_segment_t;
typedef struct gru_control_block_s gru_control_block_t;

/* Environment variables for controlling behavior*/

/*
 * Override TLBMISS fault map mode
 * 	- "user_polling", "interrupt", "os_polling"
 */
#define GRU_TLBMISS_MODE_ENV		"GRU_TLBMISS_MODE"

/* Statistics filename */
#define GRU_STATISTICS_FILE		"GRU_STATISTICS_FILE"

/* Enable mesq statistics */
#define GRU_KERNEL_MESQ_STATISTICS	"GRU_KERNEL_MESQ_STATISTICS"

/* Tarce filename */
#define GRU_TRACE_FILE			"GRU_TRACE_FILE"

/* Trace GRU instructions */
#define GRU_TRACE_INSTRUCTIONS		"GRU_TRACE_INSTRUCTIONS"

/* Trace GRU instruction TLB misses */
#define GRU_TRACE_TLB_MISS		"GRU_TRACE_TLB_MISS"

/* Trace GRU exceptions */
#define GRU_TRACE_EXCEPTIONS		"GRU_TRACE_EXCEPTIONS"

/* Trace GRU instruction retry */
#define GRU_TRACE_INSTRUCTION_RETRY	"GRU_TRACE_INSTRUCTION_RETRY"

/* Return MQE_CONGESTION for AMO_NACKED & LB_OVERFLOWED. If not
 * selected, libgru.so will handle these conditions internally.
 * 	(Experimental)
 */
#define GRU_MESQ_CONGESTION		"GRU_MESQ_CONGESTION"

/* CCH slice */
#define GRU_CCH_REQUEST_SLICE		"GRU_CCH_REQUEST_SLICE"

/* Number of TLB pages to preload into TLB on bcopy TLB miss */
#define GRU_TLB_PRELOAD			"GRU_TLB_PRELOAD"

/* Set exception retry count for numalink timeout & memory parity */
#define GRU_EXCEPTION_RETRY_ENV		"GRU_EXCEPTION_RETRY"
#define GRU_EXCEPTION_RETRY_DEFAULT	3


/*
 * Create a new GRU context
 *	cookie		- (OUT): magic identifier of the GRU segment
 *	start 		- starting address for mmaped segments (NULL means
 *			  OS picks address).
 *	ctlblks 	- number of active control blocks
 *	dataseg_bytes 	- number of data segment bytes
 *	max_threads 	- maximum number of threads that will use the context
 *	options 	- specifies various options
 *			  (see constants below)
 *
 *  Returns 0 if successful, else error code returned in errno
 */
extern int gru_create_context(gru_cookie_t *cookie, void *start,
			      unsigned int ctlblks, unsigned int dataseg_bytes,
			      unsigned int max_threads, unsigned int options);


/*
 * Destroy a GRU context
 * 	cookie	- cookie returned from gru_create_context()
 *
 * Returns:
 * 	 0 - success
 * 	-1 - failure. See errno for additional status
 */
extern int gru_destroy_context(gru_cookie_t cookie);


/*
 * Get the handle to a thread's private GRU segment
 * 	cookie		- cookie returned from gru_create_context()
 * 	threadnum	- thread number (0 .. #threads-1)
 *
 * Returns pointer to GSeg if successful, else returns NULL.
 * Error code returned in errno
 */
gru_segment_t *gru_get_thread_gru_segment(gru_cookie_t cookie, int threadnum);

/*
 * Flush a range of virtual addresses from the GRU TLB (intended for testcases
 * only)
 */
int gru_flush_tlb(gru_segment_t *gseg, void *vaddr, unsigned long len);

/*
 * Set blade and/or chiplet for GRU context. Default is to assign either chiplet
 * on the local blade.
 * 	blade   = 0 .. (num_blades - 1) (-1 implies local blade)
 * 	chiplet = (0 .. 1) (-1 implies either chiplet)
 */
int gru_set_context_blade_chiplet(gru_segment_t *gseg, int blade, int chiplet);

/*
 * Return the TLB miss handling mode
 * 	cookie		- cookie returned from gru_create_context()
 */
int gru_get_tlb_miss_mode(gru_cookie_t cookie);

/*
 * Return the TLB preload page count
 * 	cookie		- cookie returned from gru_create_context()
 */
int gru_get_tlb_preload_pages(gru_cookie_t cookie);

/*
 * Set TLB preload page count
 */
int gru_set_tlb_preload_pages(gru_segment_t *gseg, int pages);

/*
 * Unload a GRU context & free GRU resource. Will be reloaded on next
 * reference.
 */
int gru_unload_context(gru_segment_t *gseg);

/*
 * Fetch statistics for a gseg
 * 	vaddr	= any address within the gseg
 * 	stats	= pointer to statistics record
 */
int gru_get_gseg_statistics(void *vaddr, struct gru_gseg_statistics *stats);

/*
 * Determine hub type (UV1 or UV2)
 */
enum gru_hub_type {gru_uv1_hub = 2, gru_uv2_hub = 3, gru_uv3_hub = 5, gru_uv4_hub = 7};
extern enum gru_hub_type gru_hub_type;

/*
 * gru_is_uv1_hub - return TRUE iff running on a UV1 hub.
 */
static inline int gru_is_uv1_hub(void)
{
	return gru_hub_type <= gru_uv1_hub;
}

/*
 * gru_is_uv2_hub - return TRUE iff running on a UV2 hub.
 */
static inline int gru_is_uv2_hub(void)
{
	return gru_hub_type == gru_uv2_hub;
}

/*
 * gru_is_uv3_hub - return TRUE iff running on a UV3 hub.
 */
static inline int gru_is_uv3_hub(void)
{
	return gru_hub_type == gru_uv3_hub;
}

/*
 * gru_is_uv4_hub - return TRUE iff running on a UV4 hub.
 */
static inline int gru_is_uv4_hub(void)
{
	return gru_hub_type == gru_uv4_hub;
}

/*
 * Convert a user virtual address into a physical address
 * and pagesize.
 *	vaddr	  - virtual address to be translated
 *	paddr	  - if not NULL, physical address is returned
 *	pageshift - if not NULL, page-shift is returned
 *
 */
int gru_user_vtop(gru_segment_t *gseg, void *vaddr, unsigned long *paddr, int *pageshift);
#endif		/* __GRU_H__ */
