/*
 *  Copyright (c) 2008-2014 Silicon Graphics, Inc.  All Rights Reserved.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation; either version 2.1 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 */

#ifndef __GRU_INSTRUCTIONS_H__
#define __GRU_INSTRUCTIONS_H__

extern int gru_check_status_proc(gru_control_block_t *cb);
extern int gru_wait_proc(gru_control_block_t *cb);
extern void gru_wait_abort_proc(gru_control_block_t *cb);

extern void gru_abort(int, gru_control_block_t *cb, char *str);
extern char *gru_get_cb_exception_detail_str(int ret, gru_control_block_t *cb,
				char *buf, int size);


/*
 * Architecture dependent functions
 */

#define __barrier()			__asm__ __volatile__("": : :"memory")
#define __mb()				asm volatile("mfence":::"memory")
#define __rmb()				asm volatile("lfence":::"memory")
#define __wmb()				asm volatile("sfence":::"memory")
#define __flush_cache(p)		do {asm volatile("clflush %0" : "+m" (*(volatile char *)p));} while (0)
#define gru_ordered_store_ulong(p, v)	do {__barrier(); *(unsigned long*)p = v;} while (0)

/*
 * Control block status and exception codes
 */
#define CBS_IDLE			0
#define CBS_EXCEPTION			1
#define CBS_ACTIVE			2
#define CBS_CALL_OS			3

/* CB substatus bitmasks */
#define CBSS_MSG_QUEUE_MASK		7
#define CBSS_IMPLICIT_ABORT_ACTIVE_MASK	8

/* CB substatus message queue values (low 3 bits of substatus) */
#define CBSS_NO_ERROR			0
#define CBSS_LB_OVERFLOWED		1
#define CBSS_QLIMIT_REACHED		2
#define CBSS_PAGE_OVERFLOW		3
#define CBSS_AMO_NACKED			4
#define CBSS_PUT_NACKED			5

/* CB substatus message queue values for UV3 */
#define CBSS_NO_ERROR_UV3		0
#define CBSS_PUT_ERROR_FATAL_UV3	1
#define CBSS_QLIMIT_REACHED_UV3		2
#define CBSS_PAGE_OVERFLOW_UV3		3
#define CBSS_PUT_TIMEOUT_UV3		4
#define CBSS_PUT_ERROR_NONFATAL_UV3	5
#define CBSS_AMO_NACKED_UV3		6
#define CBSS_INVAL_NACKD_CODE_UV3	7

/*
 * Instruction formats
 */

/*
 * Generic instruction format.
 * This definition has precise bit field definitions.
 */
struct gru_instruction_bits {
    /* DW 0  - low */
    unsigned int		icmd:      1;
    unsigned char		ima:	   5;	/* Priority, PreFetchHint, CB_Int, Unmapped, CB_DelRep */
    unsigned char		reserved0: 2;
    unsigned int		xtype:     3;
    unsigned int		iaa0:      2;
    unsigned int		iaa1:      2;
    unsigned char		reserved1: 1;
    unsigned char		opc:       8;	/* opcode */
    unsigned char		exopc:     8;	/* extended opcode */
    /* DW 0  - high */
    unsigned int		idef2:    22;	/* TRi0 */
    unsigned char		reserved2: 2;
    unsigned char		istatus:   2;
    unsigned char		isubstatus:4;
    unsigned char		abortdet:  1;
    unsigned char		tlb_fault_color: 1;
    /* DW 1 */
    union {
	unsigned long		idef4_64;	/* 64 bits: TRi1, BufSize, Payload0 */
	struct {
	    unsigned int		idef4;	/* 32 bits: TRi1, BufSize */
	    unsigned char		sw_reserved0;
	    unsigned char		sw_reserved1;
	    unsigned short		sw_reserved2;
	};
    };
    /* DW 2-6 */
    unsigned long		idef1;		/* BAddr0 */
    unsigned long		idef5;		/* Nelem, Payload1 */
    unsigned long		idef6;		/* Stride, Operand1, Payload2 */
    unsigned long		idef3;		/* BAddr1, Value, Operand2, Payload3 */
    unsigned long		payload4;	/* Payload4 */
    /* DW 7 */
    unsigned long		avalue;		 /* AValue, Payload5 */
};

/*
 * Generic instruction with friendlier names. This format is used
 * for inline instructions.
 */
struct gru_instruction {
    /* DW 0 */
    union {
    	unsigned long		op64;    /* icmd,xtype,iaa0,ima,opc,tri0 */
	struct {
		unsigned int	op32;
		unsigned int	tri0;
	};
    };
    union {
	unsigned long		tri1_bufsize_64;	/* DW 1 */
	struct {
    		unsigned int	tri1_bufsize_32;
		unsigned char	sw_reserved0;
		unsigned char	sw_reserved1;
		unsigned short	sw_reserved2;
	};
    };
    unsigned long		baddr0;			/* DW 2 */
    unsigned long		nelem;			/* DW 3 */
    unsigned long		op1_stride;		/* DW 4 */
    unsigned long		op2_value_baddr1;	/* DW 5 */
    unsigned long		payload4;		/* DW 6 */
    unsigned long		avalue;			/* DW 7 */
};

/* Some shifts and masks for the low 64 bits of a GRU command */
#define GRU_CB_ICMD_SHFT	0
#define GRU_CB_ICMD_MASK	0x1
#define GRU_CB_XTYPE_SHFT	8
#define GRU_CB_XTYPE_MASK	0x7
#define GRU_CB_IAA0_SHFT	11
#define GRU_CB_IAA0_MASK	0x3
#define GRU_CB_IAA1_SHFT	13
#define GRU_CB_IAA1_MASK	0x3
#define GRU_CB_IMA_SHFT		1
#define GRU_CB_IMA_MASK		0x3
#define GRU_CB_OPC_SHFT		16
#define GRU_CB_OPC_MASK		0xff
#define GRU_CB_EXOPC_SHFT	24
#define GRU_CB_EXOPC_MASK	0xff
#define GRU_IDEF2_SHFT		32
#define GRU_IDEF2_MASK		0x3ffff
#define GRU_ISTATUS_SHFT	56
#define GRU_ISTATUS_MASK	0x3

/* GRU instruction opcodes (opc field) */
#define OP_NOP		0x00
#define OP_BCOPY	0x01
#define OP_VLOAD	0x02
#define OP_IVLOAD	0x03
#define OP_VSTORE	0x04
#define OP_IVSTORE	0x05
#define OP_VSET		0x06
#define OP_IVSET	0x07
#define OP_MESQ		0x08
#define OP_GAMXR	0x09
#define OP_GAMIR	0x0a
#define OP_GAMIRR	0x0b
#define OP_GAMER	0x0c
#define OP_GAMERR	0x0d
#define OP_BSTORE	0x0e
#define OP_VFLUSH	0x0f
#define OP_IVRAMIW	0x10	/* UV2 and beyond */
#define OP_IVRAMEW	0x11	/* UV2 and beyond */
#define OP_CLSTORE	0x12	/* UV3 and beyond */
#define OP_MESQ0	0x13	/* UV3 and beyond */
#define OP_DWLOAD	0x14	/* UV3 and beyond */
#define OP_DWSTORE	0x15	/* UV3 and beyond */
#define OP_GRAMER	0x16	/* UV3 and beyond */
#define OP_GRAMIR	0x17	/* UV4 and beyond */


/* Extended opcodes values (exopc field) */

/* GAMIR - AMOs with implicit operands */
#define EOP_IR_FETCH	0x01 /* Plain fetch of memory */
#define EOP_IR_CLR	0x02 /* Fetch and clear */
#define EOP_IR_INC	0x05 /* Fetch and increment */
#define EOP_IR_DEC	0x07 /* Fetch and decrement */
#define EOP_IR_QCHK1	0x0d /* Queue check, 64 byte msg */
#define EOP_IR_QCHK2	0x0e /* Queue check, 128 byte msg */

/* GAMIRR - Registered AMOs with implicit operands */
#define EOP_IRR_FETCH	0x01 /* Registered fetch of memory */
#define EOP_IRR_CLR	0x02 /* Registered fetch and clear */
#define EOP_IRR_INC	0x05 /* Registered fetch and increment */
#define EOP_IRR_DEC	0x07 /* Registered fetch and decrement */
#define EOP_IRR_DECZ	0x0f /* Registered fetch and decrement, update on zero*/

/* GAMER/GRAMER - AMOs with explicit operands */
#define EOP_ER_SWAP	0x00 /* Exchange argument and memory */
#define EOP_ER_OR	0x01 /* Logical OR with memory */
#define EOP_ER_AND	0x02 /* Logical AND with memory */
#define EOP_ER_XOR	0x03 /* Logical XOR with memory */
#define EOP_ER_ADD	0x04 /* Add value to memory */
#define EOP_ER_CSWAP	0x08 /* Compare with operand2, write operand1 if match*/
#define EOP_ER_CADD	0x0c /* 32 bit Add, write result if < limit */
#define EOP_ER_DMCAD	0x0d /* 64 bit Add with masked carry outs */

/* GAMERR - Registered AMOs with explicit operands */
#define EOP_ERR_SWAP	0x00 /* Exchange argument and memory */
#define EOP_ERR_OR	0x01 /* Logical OR with memory */
#define EOP_ERR_AND	0x02 /* Logical AND with memory */
#define EOP_ERR_XOR	0x03 /* Logical XOR with memory */
#define EOP_ERR_ADD	0x04 /* Add value to memory */
#define EOP_ERR_CSWAP	0x08 /* Compare with operand2, write operand1 if match*/
#define EOP_ERR_EPOLL	0x09 /* Poll for equality */
#define EOP_ERR_NPOLL	0x0a /* Poll for inequality */

/* IVRAMIW - AMOs with indirect implicit operands */
#define EOP_IW_AND	0x02 /* Logical AND with memory */
#define EOP_IW_INC	0x05 /* Fetch and increment */
#define EOP_IW_DEC	0x07 /* Fetch and decrement */

/* IVRAMEW - AMOs with indirect explicit operands */
#define EOP_EW_SWAP	0x00 /* Exchange argument and memory */
#define EOP_EW_OR	0x01 /* Logical OR with memory */
#define EOP_EW_AND	0x02 /* Logical AND with memory */
#define EOP_EW_XOR	0x03 /* Logical XOR with memory */
#define EOP_EW_ADD	0x04 /* Add value to memory */

/* GAMXR - SGI Arithmetic unit */
#define EOP_XR_CSWAP	0x0b /* Masked compare exchange */


/* Transfer types (xtype field) */
#define XTYPE_B		0x0	/* byte */
#define XTYPE_S		0x1	/* short (2-byte) */
#define XTYPE_W		0x2	/* word (4-byte) */
#define XTYPE_DW	0x3	/* doubleword (8-byte) */
#define XTYPE_CL	0x6	/* cacheline (64-byte) */


/* Instruction access attributes (iaa0, iaa1 fields) */
#define IAA_RAM		0x0	/* normal cached RAM access */
#define IAA_NCRAM	0x2	/* noncoherent RAM access */
#define IAA_MMIO	0x1	/* noncoherent memory-mapped I/O space */
#define IAA_REGISTER	0x3	/* memory-mapped registers, etc. */


/* Instruction mode attributes (ima field) */
#define IMA_MAPPED	0x0	/* Virtual mode  */
#define IMA_CB_DELAY	0x1	/* hold read responses until status changes */
#define IMA_UNMAPPED	0x2	/* bypass the TLBs (OS only) */
#define IMA_INTERRUPT	0x4	/* Interrupt when instruction completes */
#define IMA_PREFETCH	0x8	/* Prefetch into processor cache */
#define IMA_PRIORITY	0x10	/* Preempt lower priority instruction */

/*
 * Used for special kernel handling of mesq instructions.See mesq_internals
 * for details but basically if any IMA software bits are set, then new
 * GRU instructions preserve bit 0 of sw_reserved0 and put the IMA software
 * bits [7:4] into sw_reserved bits {4:1]. Note: hardware use bits [2:0]
 * of IMA, software uses [7:3])
 */
#define IMA_PRESERVE_TRI1_MASK	0x100000000UL	/* These sw_reserved0 bits are preserved */
#define IMA_SW_SHIFT		5		/* IMA shift to isolate software bits */
#define IMA_SW_TRI1_SHIFT	32		/* Location for IMA SW bits in TRI1 DW */

/* CBE ecause bits */
#define CBE_CAUSE_RI				(1 << 0)
#define CBE_CAUSE_INVALID_INSTRUCTION		(1 << 1)
#define CBE_CAUSE_UNMAPPED_MODE_FORBIDDEN	(1 << 2)
#define CBE_CAUSE_PE_CHECK_DATA_ERROR		(1 << 3)
#define CBE_CAUSE_IAA_GAA_MISMATCH		(1 << 4)
#define CBE_CAUSE_DATA_SEGMENT_LIMIT_EXCEPTION	(1 << 5)
#define CBE_CAUSE_OS_FATAL_TLB_FAULT		(1 << 6)
#define CBE_CAUSE_EXECUTION_HW_ERROR		(1 << 7)
#define CBE_CAUSE_TLBHW_ERROR			(1 << 8)
#define CBE_CAUSE_RA_REQUEST_TIMEOUT		(1 << 9)
#define CBE_CAUSE_HA_REQUEST_TIMEOUT		(1 << 10)
#define CBE_CAUSE_RA_RESPONSE_FATAL		(1 << 11)
#define CBE_CAUSE_RA_RESPONSE_NON_FATAL		(1 << 12)
#define CBE_CAUSE_HA_RESPONSE_FATAL		(1 << 13)
#define CBE_CAUSE_HA_RESPONSE_NON_FATAL		(1 << 14)
#define CBE_CAUSE_ADDRESS_SPACE_DECODE_ERROR	(1 << 15)
#define CBE_CAUSE_PROTOCOL_STATE_DATA_ERROR	(1 << 16)
#define CBE_CAUSE_RA_RESPONSE_DATA_ERROR	(1 << 17)
#define CBE_CAUSE_HA_RESPONSE_DATA_ERROR	(1 << 18)
#define CBE_CAUSE_FORCED_ERROR			(1 << 19)

/* CBE cbrexecstatus bits */
#define CBR_EXS_ABORT_OCC_BIT			0
#define CBR_EXS_INT_OCC_BIT			1
#define CBR_EXS_PENDING_BIT			2
#define CBR_EXS_QUEUED_BIT			3
#define CBR_EXS_TLB_INVAL_BIT			4
#define CBR_EXS_EXCEPTION_BIT			5
#define CBR_EXS_CB_INT_PENDING_BIT		6

#define CBR_EXS_ABORT_OCC			(1 << CBR_EXS_ABORT_OCC_BIT)
#define CBR_EXS_INT_OCC				(1 << CBR_EXS_INT_OCC_BIT)
#define CBR_EXS_PENDING				(1 << CBR_EXS_PENDING_BIT)
#define CBR_EXS_QUEUED				(1 << CBR_EXS_QUEUED_BIT)
#define CBR_EXS_TLB_INVAL			(1 << CBR_EXS_TLB_INVAL_BIT)
#define CBR_EXS_EXCEPTION			(1 << CBR_EXS_EXCEPTION_BIT)
#define CBR_EXS_CB_INT_PENDING			(1 << CBR_EXS_CB_INT_PENDING_BIT)

/* Message queue head structure */
union gru_mesqhead {
	unsigned long	val;
	struct {
		unsigned int	head;
		unsigned int	limit;
	};
};

/* Generate the low word of a GRU instruction */
static inline unsigned long
__opdword(struct gru_instruction *ins, unsigned char opcode,
       unsigned char exopc, unsigned char xtype,
       unsigned char iaa0, unsigned char iaa1,
       unsigned long idef2, unsigned char ima)
{
    unsigned long dw;

    dw =   (1 << GRU_CB_ICMD_SHFT) |
	   ((unsigned long)CBS_ACTIVE << GRU_ISTATUS_SHFT) |
	   (idef2<< GRU_IDEF2_SHFT) |
	   (iaa0 << GRU_CB_IAA0_SHFT) |
	   (iaa1 << GRU_CB_IAA1_SHFT) |
	   (ima << GRU_CB_IMA_SHFT) |
	   (IMA_MAPPED << GRU_CB_IMA_SHFT) |
	   (xtype << GRU_CB_XTYPE_SHFT) |
	   (opcode << GRU_CB_OPC_SHFT) |
	   (exopc << GRU_CB_EXOPC_SHFT);
    return dw;
}

/*
 * Architecture specific intrinsics
 */
static inline void gru_flush_cache(void *p)
{
	__flush_cache(p);
}

/*
 * Store the lower 64 bits of the command including the "start" bit. Then
 * start the instruction executing.
 */
static inline void gru_start_instruction(struct gru_instruction *ins, unsigned long op64)
{
	gru_ordered_store_ulong(ins, op64);
	__mb();
	gru_flush_cache(ins);
}

static inline void gru_set_tri1_value(struct gru_instruction *ins, unsigned long tri1,
		int ima)
{
	if (ima >> IMA_SW_SHIFT) {
		tri1 |= (ins->tri1_bufsize_64 & IMA_PRESERVE_TRI1_MASK);
		tri1 |= ((unsigned long)((ima >> IMA_SW_SHIFT) & ~1) << IMA_SW_TRI1_SHIFT);
	}
	ins->tri1_bufsize_64 = tri1;
}

/* Inline functions for GRU instructions.
 *     Note:
 *     	- nelem and stride are in elements
 *     	- tri0/tri1 is in bytes for the beginning of the data segment.
 */
static inline void gru_vload(gru_control_block_t *cb, void *src,
		unsigned int tri0, unsigned char xtype, unsigned long nelem,
		unsigned long stride, unsigned long hints)
{
	struct gru_instruction *ins = (struct gru_instruction *)cb;

	ins->baddr0 = (long)src;
	ins->nelem = nelem;
	ins->op1_stride = stride;
	gru_set_tri1_value(ins, 0, hints);
	gru_start_instruction(ins, __opdword(ins, OP_VLOAD, 0, xtype, IAA_RAM, 0,
					(unsigned long)tri0, hints));
}

static inline void gru_vstore(gru_control_block_t *cb, void *dest,
		unsigned int tri0, unsigned char xtype, unsigned long nelem,
		unsigned long stride, unsigned long hints)
{
	struct gru_instruction *ins = (void *)cb;

	ins->baddr0 = (long)dest;
	ins->nelem = nelem;
	ins->op1_stride = stride;
	gru_set_tri1_value(ins, 0, hints);
	gru_start_instruction(ins, __opdword(ins, OP_VSTORE, 0, xtype, IAA_RAM, 0,
					tri0, hints));
}

static inline void gru_ivload(gru_control_block_t *cb, void *src,
		unsigned int tri0, unsigned int tri1, unsigned char xtype,
		unsigned long nelem, unsigned long hints)
{
	struct gru_instruction *ins = (void *)cb;

	ins->baddr0 = (long)src;
	ins->nelem = nelem;
	gru_set_tri1_value(ins, tri1, hints);
	gru_start_instruction(ins, __opdword(ins, OP_IVLOAD, 0, xtype, IAA_RAM, 0,
					tri0, hints));
}

static inline void gru_ivstore(gru_control_block_t *cb, void *dest,
		unsigned int tri0, unsigned int tri1,
		unsigned char xtype, unsigned long nelem, unsigned long hints)
{
	struct gru_instruction *ins = (void *)cb;

	ins->baddr0 = (long)dest;
	ins->nelem = nelem;
	gru_set_tri1_value(ins, tri1, hints);
	gru_start_instruction(ins, __opdword(ins, OP_IVSTORE, 0, xtype, IAA_RAM, 0,
					tri0, hints));
}

static inline void gru_dwload(gru_control_block_t *cb, const void *src,
		unsigned char xtype, unsigned long hints)
{
	struct gru_instruction *ins = (struct gru_instruction *)cb;

	ins->baddr0 = (long)src;
	gru_start_instruction(ins, __opdword(ins, OP_DWLOAD, 0, xtype, IAA_RAM,
					0, 0, hints));
}

static inline void gru_dwstore(gru_control_block_t *cb, void *dest,
	unsigned char xtype, unsigned long hints)
{
	struct gru_instruction *ins = (void *)cb;

	ins->baddr0 = (long)dest;
	gru_start_instruction(ins, __opdword(ins, OP_DWSTORE, 0, xtype, IAA_RAM,
					0, 0, hints));
}

static inline void gru_clstore(gru_control_block_t *cb, void *dest,
		unsigned long hints)
{
	struct gru_instruction *ins = (void *)cb;

	ins->baddr0 = (long)dest;
	gru_start_instruction(ins, __opdword(ins, OP_CLSTORE, 0, XTYPE_CL,
					IAA_RAM, 0, 0, hints));
}

static inline void gru_vset(gru_control_block_t *cb, void *dest,
		unsigned long value, unsigned char xtype, unsigned long nelem,
		unsigned long stride, unsigned long hints)
{
	struct gru_instruction *ins = (void *)cb;

	ins->baddr0 = (long)dest;
	ins->op2_value_baddr1 = value;
	ins->nelem = nelem;
	gru_set_tri1_value(ins, 0, hints);
	ins->op1_stride = stride;
	gru_start_instruction(ins, __opdword(ins, OP_VSET, 0, xtype, IAA_RAM, 0,
					 0, hints));
}

extern void __gru_check_ivset_war(gru_control_block_t *cb, int xtype) __attribute__ ((weak));
static inline void gru_ivset(gru_control_block_t *cb, void *dest,
		unsigned int tri1, unsigned long value, unsigned char xtype,
		unsigned long nelem, unsigned long hints)
{
	struct gru_instruction *ins = (void *)cb;

	ins->baddr0 = (long)dest;
	ins->op2_value_baddr1 = value;
	ins->nelem = nelem;
	gru_set_tri1_value(ins, tri1, hints);
	/*
	 * UV2_REV_1_WARS - nelem > 512 will fail on UV2.
	 * The library will set nelem = 0 to cause an exception. The man page
	 * for ivset has additional info about this bug.
	 */
	if (nelem > 512 && __gru_check_ivset_war)
		__gru_check_ivset_war(cb, xtype);
	gru_start_instruction(ins, __opdword(ins, OP_IVSET, 0, xtype, IAA_RAM, 0,
					0, hints));
}

static inline void gru_vflush(gru_control_block_t *cb, void *addr,
		unsigned long nelem, unsigned char xtype, unsigned long stride,
		unsigned long hints)
{
	struct gru_instruction *ins = (void *)cb;

	ins->baddr0 = (long)addr;
	ins->op1_stride = stride;
	ins->nelem = nelem;
	gru_set_tri1_value(ins, 0, hints);
	gru_start_instruction(ins, __opdword(ins, OP_VFLUSH, 0, xtype, IAA_RAM, 0,
					0, hints));
}

static inline void gru_nop(gru_control_block_t *cb, int hints)
{
	struct gru_instruction *ins = (void *)cb;

	gru_set_tri1_value(ins, 0, hints);
	gru_start_instruction(ins, __opdword(ins, OP_NOP, 0, 0, 0, 0, 0, hints));
}


static inline void gru_bcopy(gru_control_block_t *cb, const void *src,
		void *dest,
		unsigned int tri0, unsigned int xtype, unsigned long nelem,
		unsigned int bufsize, unsigned long hints)
{
	struct gru_instruction *ins = (void *)cb;

	ins->baddr0 = (long)src;
	ins->op2_value_baddr1 = (long)dest;
	ins->nelem = nelem;
	gru_set_tri1_value(ins, bufsize, hints);
	if (bufsize == 0)
		tri0 = 0;
	gru_start_instruction(ins, __opdword(ins, OP_BCOPY, 0, xtype, IAA_RAM,
					IAA_RAM, tri0, hints));
}

static inline void gru_bstore(gru_control_block_t *cb, const void *src,
		void *dest, unsigned int tri0, unsigned int xtype,
		unsigned long nelem, unsigned long hints)
{
	struct gru_instruction *ins = (void *)cb;

	ins->baddr0 = (long)src;
	ins->op2_value_baddr1 = (long)dest;
	ins->nelem = nelem;
	gru_set_tri1_value(ins, 0, hints);
	gru_start_instruction(ins, __opdword(ins, OP_BSTORE, 0, xtype, 0, IAA_RAM,
					tri0, hints));
}

static inline void gru_gamir(gru_control_block_t *cb, int exopc, void *addr,
		unsigned int xtype, unsigned long hints)
{
	struct gru_instruction *ins = (void *)cb;

	ins->baddr0 = (long)addr;
	gru_set_tri1_value(ins, 0, hints);
	gru_start_instruction(ins, __opdword(ins, OP_GAMIR, exopc, xtype, IAA_RAM, 0,
					0, hints));
}

static inline void gru_gamirr(gru_control_block_t *cb, int exopc, void *addr,
		unsigned int xtype, unsigned long hints)
{
	struct gru_instruction *ins = (void *)cb;

	ins->baddr0 = (long)addr;
	gru_set_tri1_value(ins, 0, hints);
	gru_start_instruction(ins, __opdword(ins, OP_GAMIRR, exopc, xtype, IAA_RAM, 0,
					0, hints));
}

static inline void gru_gramir(gru_control_block_t *cb, int exopc, void *addr,
		unsigned int xtype, unsigned long hints)
{
	struct gru_instruction *ins = (void *)cb;

	ins->baddr0 = (long)addr;
	gru_set_tri1_value(ins, 0, hints);
	if (exopc & 0xf8)
		exopc = 0xff;
	gru_start_instruction(ins, __opdword(ins, OP_GRAMIR, exopc, xtype, IAA_RAM, 0,
					0, hints));
}

static inline void gru_gamer(gru_control_block_t *cb, int exopc, void *addr,
		unsigned int xtype,
		unsigned long operand1, unsigned long operand2,
		unsigned long hints)
{
	struct gru_instruction *ins = (void *)cb;

	ins->baddr0 = (long)addr;
	ins->op1_stride = operand1;
	ins->op2_value_baddr1 = operand2;
	gru_set_tri1_value(ins, 0, hints);
	gru_start_instruction(ins, __opdword(ins, OP_GAMER, exopc, xtype, IAA_RAM, 0,
					0, hints));
}

static inline void gru_gramer(gru_control_block_t *cb, int exopc,
		void *addr,
		unsigned int xtype,
		unsigned long operand1, unsigned long operand2,
		unsigned long hints)
{
	struct gru_instruction *ins = (void *)cb;

	ins->baddr0 = (long)addr;
	ins->op1_stride = operand1;
	ins->op2_value_baddr1 = operand2;
	gru_set_tri1_value(ins, 0, hints);
	gru_start_instruction(ins, __opdword(ins, OP_GRAMER, exopc, xtype, IAA_RAM, 0,
					0, hints));
}

/*
 * EOP_ERR_EPOLL & EOP_ERR_NPOLL are invalid values for exopc on UV1/2 hubs.
 * The normal gru_gamerr() function includes a hub-specific test of hub type &
 * forces an error on UV1/2 hubs. This test adds extra code. If the function is
 * known to be called only on > UV2 hubs, the extra overhead can be avoided by
 * calling __gru_gamerr directly.
 */
static inline void __gru_gamerr(gru_control_block_t *cb, int exopc, void *addr,
		unsigned int xtype, unsigned long operand1,
		unsigned long operand2, unsigned long hints)
{
	struct gru_instruction *ins = (void *)cb;

	ins->baddr0 = (long)addr;
	ins->op1_stride = operand1;
	ins->op2_value_baddr1 = operand2;
	gru_set_tri1_value(ins, 0, hints);
	gru_start_instruction(ins, __opdword(ins, OP_GAMERR, exopc, xtype, IAA_RAM, 0,
					0, hints));
}

static inline void gru_gamerr(gru_control_block_t *cb, int exopc, void *addr,
		unsigned int xtype, unsigned long operand1,
		unsigned long operand2, unsigned long hints)
{
	if ((gru_is_uv1_hub() || gru_is_uv2_hub()) &&
			(exopc == EOP_ERR_EPOLL || exopc == EOP_ERR_NPOLL))
		exopc |= 0x80;
	__gru_gamerr(cb, exopc, addr, xtype, operand1, operand2, hints);
}

static inline void gru_gamxr(gru_control_block_t *cb, void *addr,
		unsigned int tri0, unsigned long hints)
{
	struct gru_instruction *ins = (void *)cb;

	ins->baddr0 = (long)addr;
	ins->nelem = 4;
	gru_set_tri1_value(ins, 0, hints);
	gru_start_instruction(ins, __opdword(ins, OP_GAMXR, EOP_XR_CSWAP,
				 XTYPE_DW, IAA_RAM, 0, tri0, hints));
}

static inline void gru_ivramiw(gru_control_block_t *cb, int exopc, void *addr,
		unsigned int tri1, unsigned int nelem, unsigned long hints)
{
	struct gru_instruction *ins = (void *)cb;

	ins->baddr0 = (long)addr;
	ins->nelem = nelem;
	gru_set_tri1_value(ins, tri1, hints);
	gru_start_instruction(ins, __opdword(ins, OP_IVRAMIW, exopc, XTYPE_DW,
				 IAA_RAM, 0, 0, hints));
}

static inline void gru_ivramew(gru_control_block_t *cb, int exopc, void *addr,
		unsigned long tri0, unsigned int tri1,
		unsigned int nelem, unsigned long hints)
{
	struct gru_instruction *ins = (void *)cb;

	ins->baddr0 = (long)addr;
	ins->nelem = nelem;
	gru_set_tri1_value(ins, tri1, hints);
	gru_start_instruction(ins, __opdword(ins, OP_IVRAMEW, exopc, XTYPE_DW,
				 IAA_RAM, 0, tri0, hints));
}

static inline void gru_mesq(gru_control_block_t *cb, void *queue,
		unsigned long tri0, unsigned long nelem,
		unsigned long hints)
{
	struct gru_instruction *ins = (void *)cb;

	ins->baddr0 = (long)queue;
	ins->nelem = nelem;
	gru_set_tri1_value(ins, 0, hints);
	gru_start_instruction(ins, __opdword(ins, OP_MESQ, 0, XTYPE_CL, IAA_RAM, 0,
					tri0, hints));
}

static inline void gru_mesq0(gru_control_block_t *cb, void *queue,
		unsigned long hints)
{
	struct gru_instruction *ins = (void *)cb;

	ins->baddr0 = (long)queue;
	gru_start_instruction(ins, __opdword(ins, OP_MESQ0, 0, XTYPE_CL,
					IAA_RAM, 0, 0, hints));
}

static inline unsigned long gru_get_amo_value(gru_control_block_t *cb)
{
	struct gru_instruction *ins = (void *)cb;

	return ins->avalue;
}

static inline int gru_get_amo_value_head(gru_control_block_t *cb)
{
	struct gru_instruction *ins = (void *)cb;

	return ins->avalue & 0xffffffff;
}

static inline int gru_get_amo_value_limit(gru_control_block_t *cb)
{
	struct gru_instruction *ins = (void *)cb;

	return ins->avalue >> 32;
}

static inline unsigned long gru_get_dw_value(gru_control_block_t *cb)
{
	struct gru_instruction *ins = (void *)cb;

	return ins->avalue;
}

static inline void gru_put_dw_value(gru_control_block_t *cb, unsigned long val)
{
	struct gru_instruction *ins = (void *)cb;

	ins->op2_value_baddr1 = val;
}

static inline void gru_put_cl_value(gru_control_block_t *cb,
		const void *src)
{
	struct gru_instruction *ins = (void *)cb;
	unsigned long *cl = (void *)src;

	ins->tri1_bufsize_64 = *cl;
	ins->nelem = *(cl + 1);
	ins->op1_stride = *(cl + 2);
	ins->op2_value_baddr1 = *(cl + 3);
	ins->payload4 = *(cl + 4);
	ins->avalue = *(cl + 5);
}

static inline union gru_mesqhead  gru_mesq_head(int head, int limit)
{
	union gru_mesqhead mqh;

	mqh.head = head;
	mqh.limit = limit;
	return mqh;
}

#define GRU_EXC_STR_SIZE		1024


/*
 * Control block definition for checking status
 */
struct gru_control_block_status {
	unsigned int	icmd		:1;
	unsigned int	ima		:5;
	unsigned int	reserved0	:2;
	unsigned int	xtype		:3;
	unsigned int	iaa0		:2;
	unsigned int	iaa1		:2;
	unsigned char	reserved1	:1;
	unsigned char	opc		:8;
	unsigned char	exopc		:8;
	unsigned int	idef2		:22;
	unsigned char	reserved2	:2;
	unsigned int	istatus		:2;
	unsigned int	isubstatus	:4;
	unsigned char	abortdet	:1;
	unsigned char	tlb_fault_color	:1;
};

/* Get CB status */
static inline int gru_get_cb_status(gru_control_block_t *cb)
{
	struct gru_control_block_status *cbs = (void *)cb;

	return cbs->istatus;
}

/* Get CB message queue substatus */
static inline int gru_get_cb_message_queue_substatus(gru_control_block_t *cb)
{
	struct gru_control_block_status *cbs = (void *)cb;

	return cbs->isubstatus & CBSS_MSG_QUEUE_MASK;
}

/* Get CB substatus */
static inline int gru_get_cb_substatus(gru_control_block_t *cb)
{
	struct gru_control_block_status *cbs = (void *)cb;

	return cbs->isubstatus;
}

/*
 * User interface to check an instruction status. UPM and exceptions
 * are handled automatically. However, this function does NOT wait
 * for an active instruction to complete.
 *
 */
static inline int gru_check_status(gru_control_block_t *cb)
{
	return gru_check_status_proc(cb);
}

/*
 * User interface (via inline function) to wait for an instruction
 * to complete. Completion status (IDLE or EXCEPTION is returned
 * to the user. Exception due to hardware errors are automatically
 * retried before returning an exception.
 *
 */
static inline int gru_wait(gru_control_block_t *cb)
{
	return gru_wait_proc(cb);
}

/*
 * Wait for CB to complete. Aborts program if error. (Note: error does NOT
 * mean TLB mis - only fatal errors such as memory parity error or user
 * bugs will cause termination.
 */
static inline void gru_wait_abort(gru_control_block_t *cb)
{
	gru_wait_abort_proc(cb);
}

/*
 * Get a pointer to a control block
 * 	gseg	- GSeg address returned from gru_get_thread_gru_segment()
 * 	index	- index of desired CB
 */
static inline gru_control_block_t *gru_get_cb_pointer(gru_segment_t *gseg,
						      int index)
{
	return (void *)gseg + GRU_CB_BASE + index * GRU_HANDLE_STRIDE;
}

/*
 * Get a pointer to a cacheline in the data segment portion of a GSeg
 * 	gseg	- GSeg address returned from gru_get_thread_gru_segment()
 * 	index	- index of desired cache line
 */
static inline void *gru_get_data_pointer(gru_segment_t *gseg, int index)
{
	return (void *)gseg + GRU_DS_BASE + index * GRU_CACHE_LINE_BYTES;
}

/*
 * Convert a vaddr into the tri index within the GSEG
 * 	vaddr		- virtual address of within gseg
 */
static inline int gru_get_tri(void *vaddr)
{
	return ((unsigned long)vaddr & (GRU_MIN_GSEG_PAGESIZE - 1)) - GRU_DS_BASE;
}

/*
 * Decode and print a GRU instruction.
 */
void gru_print_cb_detail(const char *id, int ret, void *cb);

#endif		/* __GRU_INSTRUCTIONS_H__ */
