/*
 *  Copyright (c) 2008-2013 Silicon Graphics, Inc.  All Rights Reserved.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation; either version 2.1 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 */
#ifndef __GRU_MQ_H_
#define __GRU_MQ_H_


/*
 * Message queues using the GRU to send/receive messages.
 *
 * These function allow the user to create a message queue for
 * sending/receiving 1 or 2 cacheline messages using the GRU.
 *
 * Processes SENDING messages must have an active GRU context
 * and pass a CB to the "send" function. This CB is used to send the
 * message. The CB is free when the "send" function completes.
 *
 * The receiver does not need any GRU resources.
 *
 * The functions support:
 * 	- single receiver receiving from a partition-local
 * 	  message queue.
 * 	- multiple senders. Queue may be located in another
 * 	  partition's memory.
 *
 *
 * Missing features:
 * 	- variant of the receive function that copies the
 * 	  message to a user buffer
 * 	- user options for dealing with timeouts, queue full, etc.
 * 	- statistics
 * 	- multiple receivers
 * 	- option to WAIT in the interface if no message is available
 * 	- options for handling FULL message queue
 */

typedef void gru_message_queue_t;


/*
 * Allocate a new message queue OR initialize a user allocated
 * chunk of memory to be used as a message queue..
 *
 *  Input:
 * 	qlines    number of 64-byte cache lines in the queue
 * 	bytes	  size of message queue in bytesCreate
 * 	p	  pointer to user allocated memory. Must be
 * 		  cacheline aligned. Cannot cross a page boundary
 * 		  and must be a valid socket-accessible virtual
 * 		  address. If NULL, the library allocates a queue
 * 		  using valloc().
 *
 *  Output
 *  	mq	  message queue (NULL if error - see errno)
 *
 *  Errors:
 *  	E2BIG	  Queue too big or crosses page boundary
 *  	EINVAL    <p> not cacheline aligned
 */
gru_message_queue_t *gru_create_message_queue(void *p, unsigned int bytes);


/*
 * Destroy a message queue & free resources. Use this to
 * free resource ONLY for message queues allocated by
 * gru_create_message_queue().
 *
 *   Input:
 *   	mq	  message queue. This is the address returned
 *   		  from gru_create_message_queue().
 *
 *   Output:
 *      0	  Successfully deleted
 *     -1	  Error - see errno
 *
 *   Errors:
 *      EINVAL	  mq not a message queue
 */
int gru_destroy_message_queue(gru_message_queue_t *mq);


/*
 * Send a message to a message queue.
 *
 * Note: The message queue transport mechanism uses the first 32
 * bits of the message. Users should avoid using these bits.
 *
 *   Input:
 * 	mq	message queue. This can be a partition-local
 * 		virtual address OR an exported cross-partition
 * 		address of a message queue (GRU-addressible).
 * 	cb	GRU control block to use to send message
 * 	mesg	message.  Must be vaddr within a GSEG
 * 	bytes	size of message in bytes
 *
 *   Output:
 *      0	message sent
 *     >0	Send failure - see error codes below
 *     -1	System error - see errno
 *
 */
int gru_send_message(gru_message_queue_t *mq, gru_control_block_t *cb,
		     void *mesg, unsigned int bytes);


/*
 * Send a 48 byte message to a message queue via the control block.
 *
 * Note: The message queue transport mechanism uses the first 32
 * bits of the message. Users should avoid using these bits.
 *
 *   Input:
 * 	mq	message queue. This can be a partition-local
 * 		virtual address OR an exported cross-partition
 * 		address of a message queue (GRU-addressible).
 * 	cb	GRU control block to use to send message
 * 	mesg	message.  Does not need to exist within a GSEG.
 *
 *   Output:
 *      0	message sent
 *     >0	Send failure - see error codes below
 *     -1	System error - see errno
 *
 */
int gru_send_cb_message(gru_message_queue_t *mq, gru_control_block_t *cb,
		     void *mesg);


/*
 * Initiate the sending of a message to a message queue
 *
 * Note: The message queue transport mechanism uses the first 32
 * bits of the message. Users should avoid using these bits.
 *
 *   Input:
 * 	mq	message queue. This can be a partition-local
 * 		virtual address OR an exported cross-partition
 * 		address of a message queue (GRU-addressible).
 * 	cb	GRU control block to use to send message
 * 	mesg	message.  Must be vaddr within a GSEG
 * 	bytes	size of message in bytes
 *
 *   Output:
 *      0	message send initiated
 *     >0	Send failure - see error codes below
 *     -1	System error - see errno
 *
 */
int gru_start_message(gru_message_queue_t *mq, gru_control_block_t *cb,
		     void *mesg, unsigned int bytes);


/*
 * Initiate the sending of a 48 byte message via the cb to a message queue
 *
 * Note: The message queue transport mechanism uses the first 32
 * bits of the message. Users should avoid using these bits.
 *
 *   Input:
 * 	mq	message queue. This can be a partition-local
 * 		virtual address OR an exported cross-partition
 * 		address of a message queue (GRU-addressible).
 * 	cb	GRU control block to use to send message
 * 	mesg	message.  Does not need to exist within a GSEG.
 *
 *   Output:
 *      0	message send initiated
 *     >0	Send failure - see error codes below
 *     -1	System error - see errno
 *
 */
int gru_start_cb_message(gru_message_queue_t *mq, gru_control_block_t *cb,
		     void *mesg);


/*
 * Wait for a message previously started with gru_start_message to complete
 *
 *   Input:
 * 	mq	message queue. This can be a partition-local
 * 		virtual address OR an exported cross-partition
 * 		address of a message queue (GRU-addressible).
 * 	cb	GRU control block used to send message
 * 	mesg	message.  Must be vaddr within a GSEG
 * 	bytes	size of message in bytes
 *
 *   Output:
 *      0	message delivered
 *     >0	Send failure - see error codes below
 *     -1	System error - see errno
 *
 */
int gru_wait_message(gru_message_queue_t *mq, gru_control_block_t *cb,
		     void *mesg, unsigned int bytes);


/*
 * Wait for a message previously started with gru_start_cb_message to complete
 *
 *   Input:
 * 	mq	message queue. This can be a partition-local
 * 		virtual address OR an exported cross-partition
 * 		address of a message queue (GRU-addressible).
 * 	cb	GRU control block used to send message
 * 	mesg	message.  Does not need to exist within a GSEG.
 *
 *   Output:
 *      0	message delivered
 *     >0	Send failure - see error codes below
 *     -1	System error - see errno
 *
 */
int gru_wait_cb_message(gru_message_queue_t *mq, gru_control_block_t *cb,
		     void *mesg);

/* Status values for gru_send_message() */
#define MQE_OK			0	/* message sent successfully */
#define MQE_CONGESTION		1	/* temporary failure caused by network
					   congestion, try again */
#define MQE_QUEUE_FULL		2	/* queue is full and receiver not
					   keeping up */
#define MQE_PAGE_OVERFLOW	3	/* Queue overflowed a page. User bug */
#define MQE_BUG			4	/* unexpected library bug encountered */
#define MQE_BUG_TOO_BIG		10	/* Message too big */
#define MQE_AGAIN		-2	/* (internal) queue full sucessfully
					   handled */
#define MQE_UNEXPECTED_CB_ERR	-1	/* unexpected CB error. See CB for
					   details */


/*
 * Advance the receive pointer for the message queue to the next message.
 * Note: current API requires messages to be gotten & freed in order. Future
 * API extensions may allow for out-of-order freeing.
 *
 *   Input
 * 	mq	message queue
 */
void gru_free_message(gru_message_queue_t *mq, void *mesg);


/*
 * Get next message from message queue. Returns pointer to
 * message OR NULL if no message present.
 * User must call gru_free_message() to move to next message.
 *
 *   Input
 * 	mq	message queue
 *
 *   Output:
 *	p	pointer to message
 *	NULL	no message available
 */
void *gru_get_next_message(gru_message_queue_t *mq);

#endif 		/* __GRU_MQ_H_ */
